/*
  Copyright © 2001, 2021 Jan Bobrowski

  This program is free software; you can redistribute it or
  modify it under the terms of the GNU General Public License
  version 2 as published by the Free Software Foundation.
*/

#include "toico.h"

int png_dims(u8 p[24], int *w, int *h)
{
	if (memcmp(p + 12, "IHDR", 4) != 0)
		return -1;
	*w = get32be_at(p + 16);
	*h = get32be_at(p + 20);
	return 0;
}

#ifndef WITH_PNG

int is_png(u8 p[SIGN_SIZE])
{
	u8 sign[] = {0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A};
	return memcmp(p, sign, sizeof sign) == 0;
}

#else
#include <png.h>

int is_png(u8 p[SIGN_SIZE])
{
	return png_sig_cmp(p, 0, SIGN_SIZE) == 0;
}

static void png_read(png_structp png, png_bytep b, size_t n)
{
	struct file *fi = png_get_io_ptr(png);
	read_all(fi, b, n);
}

void read_png(struct image *img, struct file *fi)
{
	png_structp png;
	png_infop pi, pe;

	png = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	if(!png)
		goto pngerr;
	pi = png_create_info_struct(png);
	if(!pi)
		goto pngerr;
	pe = png_create_info_struct(png);
	if(!pe)
		goto pngerr;
	png_set_read_fn(png, fi, png_read);
	png_read_info(png, pi);

	img->width = png_get_image_width(png, pi);
	img->height = png_get_image_height(png, pi);

	png_set_strip_16(png);
	png_set_expand(png);
	png_set_gray_to_rgb(png);
	png_set_invert_alpha(png);
	{
		const union {u16 a; u8 b[8];} u = {1};
		if (u.b[0]) {
			png_set_bgr(png);
			png_set_filler(png, 0, PNG_FILLER_AFTER);
		} else {
			png_set_filler(png, 0, PNG_FILLER_BEFORE);
		}
	}
	alloc_im(img);
	{
		png_bytep *rows = alloc(img->height * sizeof *rows);
		u32 *p = img->data;
		int y = 0;
		do {
			rows[y++] = (png_bytep)p;
			p += img->width;
		} while (y < img->height);
		png_read_image(png, rows);
		free(rows);
	}
	png_read_end(png, pe);
	png_destroy_read_struct(&png, &pi, &pe);
	return;
pngerr:
	bad("Library error");
}

#endif
