#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <err.h>
#include "list.h"

#pragma clang diagnostic ignored "-Wshift-op-parentheses"
#pragma clang diagnostic ignored "-Wbitwise-op-parentheses"
#pragma clang diagnostic ignored "-Wlogical-op-parentheses"

#define elemof(T) (sizeof T/sizeof*T)
#define endof(T) (T+elemof(T))
#define STR_(S) #S
#define STR(S) STR_(S)

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;

struct opt {
	char *file;
	unsigned verbose:1;
	unsigned mode_convert:1;
	unsigned mode_list:1;
};

extern struct opt opt;

#define SIGN_SIZE 8

struct file {
	FILE *file;
	char *name;
	u8 *ptr, *end;
	u8 buf[SIGN_SIZE];
};

enum {All, Some};
void *read_checked(struct file *fi, void *buf, int n, int all);
static inline void *read_all(struct file *fi, void *buf, int n) {return read_checked(fi, buf, n, All);}
static inline void *read_some(struct file *fi, void *buf, int n) {return read_checked(fi, buf, n, Some);}
void skip(struct file *fi, u32 n);

static inline int read_byte(struct file *fi) {
	u8 b;
	return read_some(fi, &b, 1) == &b ? -1 : b;
}

struct image {
	struct list list;
	char *name;
	int index; // in the source collection
	u32 *data;
	int width, height, depth;
	int npal;
	u8 *raw_data;
	unsigned raw_size;

	u32 alpha_key;
	unsigned has_key:1;
	unsigned no_semi:1;	/* don't generate semitransparent icons */
	unsigned embed:1;

	u32 pal[256];
};

enum {
	Compr_None = 0,
	Compr_RLE8 = 1,
	Compr_RLE4 = 2,
	Compr_Bitfields = 3,
	Compr_AlphaBitfields = 6
};

extern struct list images;
extern int nimages;
extern struct image *current;

#ifndef no_return
#define no_return
#endif
no_return void bad_image(char *fmt, ...);
#ifdef NDEBUG
#define bad(F,...) bad_image("%s: " F, current->name, ## __VA_ARGS__)
#else
#define bad(F,...) bad_image("(%s:%d) %s: " F, __FILE__, __LINE__, current->name, ## __VA_ARGS__)
#endif

static inline u32 *pixel_at(struct image *im, unsigned x, unsigned y) {return im->data + y*im->width + x;}
static inline u32 *data_end(struct image *im) {return im->data + im->width*im->height;}
static inline unsigned roundup(unsigned n, unsigned d) {return n*d + 31u >> 3 & ~3;}

#define ALPHA 0xFF000000u
static inline int transparent(u32 c) {return (c & ALPHA) == ALPHA;}
void *alloc(unsigned size);
void alloc_im(struct image *);

int is_ico(u8[SIGN_SIZE]);
int is_png(u8[SIGN_SIZE]);
int is_gif(u8[SIGN_SIZE]);
int is_tiff(u8[SIGN_SIZE]);
int is_bmp(u8[SIGN_SIZE]);
int is_xpm(u8[SIGN_SIZE]);
int is_icns(u8[SIGN_SIZE]);

int png_dims(u8 p[24], int *w, int *h);

void read_png(struct image *, struct file *);
void read_gif(struct image *, struct file *);
void read_tiff(struct image *, struct file *);
void read_xpm(struct image *, struct file *);
void read_bmp(struct image *, struct file *);
void read_icns(struct image *, struct file *);

void list_ico(struct file *);
void list_icns(struct file *);

void write_ico(FILE *);

typedef struct u16le {u8 a, b;} u16le;
typedef struct u32le {u8 a, b, c, d;} u32le;
typedef struct u16be {u8 b, a;} u16be;
typedef struct u32be {u8 d, c, b, a;} u32be;

static inline u16 get16le(u16le v) {return v.a | v.b << 8;}
static inline u32 get32le(u32le v) {return v.a | v.b << 8 | v.c << 16 | v.d << 24;}
static inline u32 get32be(u32be v) {return v.a | v.b << 8 | v.c << 16 | v.d << 24;}
static inline u32 get32be_at(void *b) {u8 *p = b; return p[0] << 24 | p[1] << 16 | p[2] << 8 | p[3];}
